@@ -1,14 +1,16 @@
-# SPDX-FileCopyrightText: 2012-2022  kaliko <kaliko@azylum.org>
+# -*- coding: utf-8 -*-
+# SPDX-FileCopyrightText: 2012-2024  kaliko <kaliko@azylum.org>
 # SPDX-FileCopyrightText: 2021       Wonko der Verständige <wonko@hanstool.org>
 # SPDX-FileCopyrightText: 2019       Naglis Jonaitis <naglis@mailbox.org>
 # SPDX-FileCopyrightText: 2019       Bart Van Loon <bbb@bbbart.be>
 # SPDX-FileCopyrightText: 2008-2010  J. Alexander Treuman <jat@spatialrift.net>
-# SPDX-License-Identifier: GPL-3.0-or-later
-"""python-musicpd: Python Music Player Daemon client library"""
+# SPDX-License-Identifier: LGPL-3.0-or-later
+"""Python Music Player Daemon client library"""
 
 
-import socket
+import logging
 import os
+import socket
 
 from functools import wraps
 
@@ -16,13 +18,16 @@ HELLO_PREFIX = "OK MPD "
 ERROR_PREFIX = "ACK "
 SUCCESS = "OK"
 NEXT = "list_OK"
-VERSION = '0.8.0'
+#: Module version
+VERSION = '0.9.0'
 #: Seconds before a connection attempt times out
-#: (overriden by MPD_TIMEOUT env. var.)
+#: (overriden by :envvar:`MPD_TIMEOUT` env. var.)
 CONNECTION_TIMEOUT = 30
-#: Socket timeout in second (Default is None for no timeout)
+#: Socket timeout in second > 0 (Default is :py:obj:`None` for no timeout)
 SOCKET_TIMEOUT = None
 
+log = logging.getLogger(__name__)
+
 
 def iterator_wrapper(func):
     """Decorator handling iterate option"""
@@ -44,59 +49,73 @@ def iterator_wrapper(func):
 
 
 class MPDError(Exception):
-    pass
+    """Main musicpd Exception"""
 
 
 class ConnectionError(MPDError):
-    pass
+    """Fatal Connection Error, cannot recover from it."""
 
 
 class ProtocolError(MPDError):
-    pass
+    """Fatal Protocol Error, cannot recover from it"""
 
 
 class CommandError(MPDError):
-    pass
+    """Malformed command, socket should be fine, can reuse it"""
 
 
 class CommandListError(MPDError):
-    pass
+    """"""
 
 
 class PendingCommandError(MPDError):
-    pass
+    """"""
 
 
 class IteratingError(MPDError):
-    pass
+    """"""
 
 
 class Range:
 
     def __init__(self, tpl):
         self.tpl = tpl
+        self.lower = ''
+        self.upper = ''
         self._check()
 
     def __str__(self):
-        if len(self.tpl) == 0:
-            return ':'
-        if len(self.tpl) == 1:
-            return '{0}:'.format(self.tpl[0])
-        return '{0[0]}:{0[1]}'.format(self.tpl)
+        return f'{self.lower}:{self.upper}'
 
     def __repr__(self):
-        return 'Range({0})'.format(self.tpl)
+        return f'Range({self.tpl})'
+
+    def _check_element(self, item):
+        if item is None or item == '':
+            return ''
+        try:
+            return str(int(item))
+        except (TypeError, ValueError) as err:
+            raise CommandError(f'Not an integer: "{item}"') from err
+        return item
 
     def _check(self):
         if not isinstance(self.tpl, tuple):
             raise CommandError('Wrong type, provide a tuple')
-        if len(self.tpl) not in [0, 1, 2]:
-            raise CommandError('length not in [0, 1, 2]')
-        for index in self.tpl:
-            try:
-                index = int(index)
-            except (TypeError, ValueError) as err:
-                raise CommandError('Not a tuple of int') from err
+        if len(self.tpl) == 0:
+            return
+        if len(self.tpl) == 1:
+            self.lower = self._check_element(self.tpl[0])
+            return
+        if len(self.tpl) != 2:
+            raise CommandError('Range wrong size (0, 1 or 2 allowed)')
+        self.lower = self._check_element(self.tpl[0])
+        self.upper = self._check_element(self.tpl[1])
+        if self.lower == '' and self.upper != '':
+            raise CommandError(f'Integer expected to start the range: {self.tpl}')
+        if self.upper.isdigit() and self.lower.isdigit():
+            if int(self.lower) > int(self.upper):
+                raise CommandError(f'Wrong range: {self.lower} > {self.upper}')
 
 
 class _NotConnected:
@@ -109,14 +128,15 @@ class _NotConnected:
 
 
 class MPDClient:
-    """MPDClient instance will look for ``MPD_HOST``/``MPD_PORT``/``XDG_RUNTIME_DIR`` environment
-    variables and set instance attribute ``host``, ``port`` and ``pwd``
-    accordingly. Regarding ``MPD_HOST`` format to expose password refer
-    MPD client manual :manpage:`mpc (1)`.
+    """MPDClient instance will look for :envvar:`MPD_HOST`/:envvar:`MPD_PORT`/:envvar:`XDG_RUNTIME_DIR` environment
+    variables and set instance attribute :py:attr:`host`, :py:attr:`port` and :py:obj:`pwd`
+    accordingly.
 
-    Then :py:obj:`musicpd.MPDClient.connect` will use ``host`` and ``port`` as defaults if not provided as args.
+    Then :py:obj:`musicpd.MPDClient.connect` will use :py:obj:`host` and
+    :py:obj:`port` as defaults if not provided as args.
 
-    Cf. :py:obj:`musicpd.MPDClient.connect` for details.
+    Regarding :envvar:`MPD_HOST` format to expose password refer this module
+    documentation or MPD client manual :manpage:`mpc (1)`.
 
     >>> from os import environ
     >>> environ['MPD_HOST'] = 'pass@mpdhost'
@@ -127,22 +147,29 @@ class MPDClient:
     True
     >>> cli.connect() # will use host/port as set in MPD_HOST/MPD_PORT
 
-    :ivar str host: host used with the current connection
-    :ivar str,int port: port used with the current connection
-    :ivar str pwd: password detected in ``MPD_HOST`` environment variable
+    .. note::
 
-    .. warning:: Instance attribute host/port/pwd
+      default host:
+       * use :envvar:`MPD_HOST` environment variable if set, extract password if present,
+       * else use :envvar:`XDG_RUNTIME_DIR` to looks for an existing file in ``${XDG_RUNTIME_DIR:-/run/}/mpd/socket``
+       * else set host to ``localhost``
 
-      While :py:attr:`musicpd.MPDClient().host` and
-      :py:attr:`musicpd.MPDClient().port` keep track of current connection
-      host and port, :py:attr:`musicpd.MPDClient().pwd` is set once with
+      default port:
+       * use :envvar:`MPD_PORT` environment variable is set
+       * else use ``6600``
+
+    .. warning:: **Instance attribute host/port/pwd**
+
+      While :py:attr:`musicpd.MPDClient.host` and
+      :py:attr:`musicpd.MPDClient.port` keep track of current connection
+      host and port, :py:attr:`musicpd.MPDClient.pwd` is set once with
       password extracted from environment variable.
-      Calling :py:meth:`musicpd.MPDClient().password()` with a new password
-      won't update :py:attr:`musicpd.MPDClient().pwd` value.
+      Calling MPS's password method with a new password
+      won't update :py:attr:`musicpd.MPDClient.pwd` value.
 
-      Moreover, :py:attr:`musicpd.MPDClient().pwd` is only an helper attribute
-      exposing password extracted from ``MPD_HOST`` environment variable, it
-      will not be used as default value for the :py:meth:`password` method
+      Moreover, :py:attr:`musicpd.MPDClient.pwd` is only an helper attribute
+      exposing password extracted from :envvar:`MPD_HOST` environment variable, it
+      will not be used as default value for the MPD's password command.
     """
 
     def __init__(self):
@@ -152,6 +179,10 @@ class MPDClient:
         #: Current connection timeout value, defaults to
         #: :py:obj:`CONNECTION_TIMEOUT` or env. var. ``MPD_TIMEOUT`` if provided
         self.mpd_timeout = None
+        self.mpd_version = ''
+        """Protocol version as exposed by the server as a :py:obj:`str`
+
+        .. note:: This is the version of the protocol spoken, not the real version of the daemon."""
         self._reset()
         self._commands = {
             # Status Commands
@@ -284,47 +315,58 @@ class MPDClient:
             "readmessages":       self._fetch_messages,
             "sendmessage":        self._fetch_nothing,
         }
+        #: host used with the current connection (:py:obj:`str`)
+        self.host = None
+        #: password detected in :envvar:`MPD_HOST` environment variable (:py:obj:`str`)
+        self.pwd = None
+        #: port used with the current connection (:py:obj:`int`, :py:obj:`str`)
+        self.port = None
         self._get_envvars()
 
     def _get_envvars(self):
         """
-        Retrieve MPD env. var. to overrides "localhost:6600"
-            Use MPD_HOST/MPD_PORT if set
-            else use MPD_HOST=${XDG_RUNTIME_DIR:-/run/}/mpd/socket if file exists
+        Retrieve MPD env. var. to overrides default "localhost:6600"
         """
+        # Set some defaults
         self.host = 'localhost'
-        self.pwd = None
         self.port = os.getenv('MPD_PORT', '6600')
-        if os.getenv('MPD_HOST'):
+        _host = os.getenv('MPD_HOST', '')
+        if _host:
             # If password is set: MPD_HOST=pass@host
-            if '@' in os.getenv('MPD_HOST'):
-                mpd_host_env = os.getenv('MPD_HOST').split('@', 1)
+            if '@' in _host:
+                mpd_host_env = _host.split('@', 1)
                 if mpd_host_env[0]:
                     # A password is actually set
+                    log.debug('password detected in MPD_HOST, set client pwd attribute')
                     self.pwd = mpd_host_env[0]
                     if mpd_host_env[1]:
                         self.host = mpd_host_env[1]
+                        log.debug('host detected in MPD_HOST: %s', self.host)
                 elif mpd_host_env[1]:
                     # No password set but leading @ is an abstract socket
                     self.host = '@'+mpd_host_env[1]
+                    log.debug('host detected in MPD_HOST: %s (abstract socket)', self.host)
             else:
                 # MPD_HOST is a plain host
-                self.host = os.getenv('MPD_HOST')
+                self.host = _host
+                log.debug('host detected in MPD_HOST: %s', self.host)
         else:
             # Is socket there
             xdg_runtime_dir = os.getenv('XDG_RUNTIME_DIR', '/run')
             rundir = os.path.join(xdg_runtime_dir, 'mpd/socket')
             if os.path.exists(rundir):
                 self.host = rundir
-        self.mpd_timeout = os.getenv('MPD_TIMEOUT')
-        if self.mpd_timeout and self.mpd_timeout.isdigit():
-            self.mpd_timeout = int(self.mpd_timeout)
+                log.debug('host detected in ${XDG_RUNTIME_DIR}/run: %s (unix socket)', self.host)
+        _mpd_timeout = os.getenv('MPD_TIMEOUT', '')
+        if _mpd_timeout.isdigit():
+            self.mpd_timeout = int(_mpd_timeout)
+            log.debug('timeout detected in MPD_TIMEOUT: %d', self.mpd_timeout)
         else:  # Use CONNECTION_TIMEOUT as default even if MPD_TIMEOUT carries gargage
             self.mpd_timeout = CONNECTION_TIMEOUT
 
     def __getattr__(self, attr):
         if attr == 'send_noidle':  # have send_noidle to cancel idle as well as noidle
-            return self.noidle()
+            return self.noidle
         if attr.startswith("send_"):
             command = attr.replace("send_", "", 1)
             wrapper = self._send
@@ -394,9 +436,9 @@ class MPDClient:
         parts = [command]
         for arg in args:
             if isinstance(arg, tuple):
-                parts.append('{0!s}'.format(Range(arg)))
+                parts.append(f'{Range(arg)!s}')
             else:
-                parts.append('"%s"' % escape(str(arg)))
+                parts.append(f'"{escape(str(arg))}"')
         if '\n' in ' '.join(parts):
             raise CommandError('new line found in the command!')
         self._write_line(" ".join(parts))
@@ -585,7 +627,7 @@ class MPDClient:
         self.mpd_version = line[len(HELLO_PREFIX):].strip()
 
     def _reset(self):
-        self.mpd_version = None
+        self.mpd_version = ''
         self._iterating = False
         self._pending = []
         self._command_list = None
@@ -600,10 +642,13 @@ class MPDClient:
         # abstract socket
         if path.startswith('@'):
             path = '\0'+path[1:]
-        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-        sock.settimeout(self.mpd_timeout)
-        sock.connect(path)
-        sock.settimeout(self.socket_timeout)
+        try:
+            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+            sock.settimeout(self.mpd_timeout)
+            sock.connect(path)
+            sock.settimeout(self.socket_timeout)
+        except socket.error as socket_err:
+            raise ConnectionError(socket_err) from socket_err
         return sock
 
     def _connect_tcp(self, host, port):
@@ -612,23 +657,29 @@ class MPDClient:
         except AttributeError:
             flags = 0
         err = None
-        for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC,
-                                      socket.SOCK_STREAM, socket.IPPROTO_TCP,
-                                      flags):
+        try:
+            gai = socket.getaddrinfo(host, port, socket.AF_UNSPEC,
+                                     socket.SOCK_STREAM, socket.IPPROTO_TCP,
+                                     flags)
+        except socket.error as gaierr:
+            raise ConnectionError(gaierr) from gaierr
+        for res in gai:
             af, socktype, proto, _, sa = res
             sock = None
             try:
+                log.debug('opening socket %s', sa)
                 sock = socket.socket(af, socktype, proto)
                 sock.settimeout(self.mpd_timeout)
                 sock.connect(sa)
                 sock.settimeout(self.socket_timeout)
                 return sock
             except socket.error as socket_err:
+                log.debug('opening socket %s failed: %s', sa, socket_err)
                 err = socket_err
                 if sock is not None:
                     sock.close()
         if err is not None:
-            raise ConnectionError(str(err))
+            raise ConnectionError(err)
         raise ConnectionError("getaddrinfo returns an empty list")
 
     def noidle(self):
@@ -642,29 +693,19 @@ class MPDClient:
     def connect(self, host=None, port=None):
         """Connects the MPD server
 
-        :param str host: hostname, IP or FQDN (defaults to `localhost` or socket, see below for details)
-        :param port: port number (defaults to 6600)
+        :param str host: hostname, IP or FQDN (defaults to *localhost* or socket)
+        :param port: port number (defaults to *6600*)
         :type port: str or int
 
-        The connect method honors MPD_HOST/MPD_PORT environment variables.
+        If host/port are :py:obj:`None` the socket uses :py:attr:`host`/:py:attr:`port`
+        attributes as defaults. Cf. :py:obj:`MPDClient` for the logic behind default host/port.
 
-        The underlying socket also honors MPD_TIMEOUT environment variable
+        The underlying socket also honors :envvar:`MPD_TIMEOUT` environment variable
         and defaults to :py:obj:`musicpd.CONNECTION_TIMEOUT` (connect command only).
 
         If you want to have a timeout for each command once you got connected,
         set its value in :py:obj:`MPDClient.socket_timeout` (in second) or at
         module level in :py:obj:`musicpd.SOCKET_TIMEOUT`.
-
-        .. note:: Default host/port
-
-          If host evaluate to :py:obj:`False`
-           * use ``MPD_HOST`` environment variable if set, extract password if present,
-           * else looks for an existing file in ``${XDG_RUNTIME_DIR:-/run/}/mpd/socket``
-           * else set host to ``localhost``
-
-          If port evaluate to :py:obj:`False`
-           * if ``MPD_PORT`` environment variable is set, use it for port
-           * else use ``6600``
         """
         if not host:
             host = self.host
@@ -677,8 +718,10 @@ class MPDClient:
         if self._sock is not None:
             raise ConnectionError("Already connected")
         if host[0] in ['/', '@']:
+            log.debug('Connecting unix socket %s', host)
             self._sock = self._connect_unix(host)
         else:
+            log.debug('Connecting tcp socket %s:%s (timeout: %ss)', host, port, self.mpd_timeout)
             self._sock = self._connect_tcp(host, port)
         self._rfile = self._sock.makefile("r", encoding='utf-8', errors='surrogateescape')
         self._rbfile = self._sock.makefile("rb")
@@ -688,19 +731,30 @@ class MPDClient:
         except:
             self.disconnect()
             raise
+        log.debug('Connected')
 
     @property
     def socket_timeout(self):
         """Socket timeout in second (defaults to :py:obj:`SOCKET_TIMEOUT`).
-        Use None to disable socket timout."""
+        Use :py:obj:`None` to disable socket timout.
+
+        :setter: Set the socket timeout (integer > 0)
+        :type: int or None
+        """
         return self._socket_timeout
 
     @socket_timeout.setter
     def socket_timeout(self, timeout):
-        self._socket_timeout = timeout
+        if timeout is not None:
+            if int(timeout) <= 0:
+                raise ValueError('socket_timeout expects a non zero positive integer')
+            self._socket_timeout = int(timeout)
+        else:
+            self._socket_timeout = timeout
         if getattr(self._sock, 'settimeout', False):
             self._sock.settimeout(self._socket_timeout)
 
+
     def disconnect(self):
         """Closes the MPD connection.
         The client closes the actual socket, it does not use the
